Unit MySock;

// *****************************************************************************
//                          TSOCK 2.0b
// Changed By Carlo Kok                   http://cajsoft.cjb.net/
// Originally made by Beach Dog Software  http://www.beachdogsoftware.com
//
// Bugreport: bugs_mysock@cajsoft.cjb.net
// *****************************************************************************
//
// Registering:
// Please register when you like mysock. You can register by sending an
// email with you name, companyname, country and what you think of mysock.
// To register_mysock@cajsoft.cjb.net. Registered will cost you nothing.
//
//
// Copyright (C) 1999 by Carlo Kok (ck@cajsoft.cjb.net)
// Copyright (C) 1998 by Beachdog software (www.beachdogsoftware.com)
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the author be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute
// it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented, you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and
// must not be misrepresented as being the original software.
//
// 3. This notice and the copyright string may not be removed
// or altered from any source distribution.
{
Changes for version 2.0a:
  - Created seperate components for client and server.
  - Removed some procedures/functions/properties
  - When turning TCustomServersock.active to false, function returns
    until client has stopped.
  - Added maxClient property, so you can set the maximum number of clients.
    Use 0 for infinite number of clients.
Version 2.0b
  - Removed several bugs
}
{$IFDEF VER100}
{$DEFINE VER3ORABOVE}
{$ENDIF}
{$IFDEF VER120}
{$DEFINE VER3ORABOVE}
{$ENDIF}

Interface

Uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
     WinSock;

Const
  Copyright : STring = 'TSock 2.0b Copyright (C) 1999 by Carlo Kok (ck@cajsoft.cjb.net) http://cajsoft.cjb.net; originally by Beachdog Software';

Const WM_SOCK = WM_USER + 72;

Type TClientSockServ = Class;
     TCustomServerSock = Class;

     ESockException = Class(Exception);
     TNotifyReadEvent = Procedure(Sender : TObject; Count : Integer) Of Object;
     TNotifyAutoEvent = Procedure(Sender : TObject; NewSock : TClientSockServ) Of Object;

     // TSock for creating socket window
     TSock = class(TComponent)
     private
       HWND : THandle;
     protected
       property Handle : THandle read HWND;
       Procedure OnSockMessage(Var Message : TMessage);virtual;abstract;
     public
       procedure ProcessMessages;
       Constructor Create(AOwner : TComponent); Override;
       Destructor Destroy; Override;
     end;

     TClientSock = Class(TSock)
        Private
           FSockAddrIn : TSockAddrIn;  // Address Information Block
           FRecvAddrIn : TSockAddrIn;  // Address Information Block For RecvFrom

           // Character Buffer (Most WINSOCK.DLLs Max At 32k)
           FCharBuf    : Array[1..32768] Of Char;
           FHostName   : String;       // Host Name Or IP Address
           FPortName   : String;       // Port Name Or Well-Known Number
           FSocket     : TSocket;      // Socket Handle
           FInBuffer   : String;       // Input Buffer
           FOutBuffer  : String;       // Output Buffer For Non-Blocking
           FBlocking   : Boolean;      // Do Blocking Calls?
           FConnected  : Boolean;      // Are We Connected?
           FBlockTime  : Integer;      // How Long To Wait For Blocking Operation
           FOnConnect    : TNotifyEvent;
           FOnDisconnect : TNotifyEvent;
           FOnRead       : TNotifyReadEvent;
           FOnWrite      : TNotifyEvent;
           // Property Set/Get Routines
           Procedure SetHostName(Value : String);
           Procedure SetPortName(Value : String);
           Function GetText : String;
           Procedure SetText(Value : String);
           Procedure SetBlocking(Value : Boolean);
           Procedure SetConnected(Value : Boolean);
           Function GetConnected : Boolean;
           Function GetRemoteHost : String;
        Protected
           Procedure OnSockMessage(Var Message : TMessage);override;
           Procedure Loaded; Override;
           Constructor CreateWithSocket(AOwner : TComponent; NewSocket : TSocket); Virtual;
        Public
           Constructor Create(AOwner : TComponent); Override;
           Destructor Destroy; Override;
           Function Open : Boolean;
           Procedure Abort;
           Function Close : Boolean;
           Function Send(Value : String) : Boolean;
           Function ReceiveCount(Count : Integer) : String;
           Function Receive : String;
           Function HostLookup(Value : String) : TInAddr;
           Function PortLookup(Value : String) : U_Short;
           Property Text : String Read GetText Write SetText;
           Property Connected : Boolean Read GetConnected Write SetConnected; // Used To Read FConnected
           Property Socket : TSocket Read FSocket;
           Property RemoteHost : String Read GetRemoteHost;
        Published
           Property HostName : String Read FHostName Write SetHostName;
           Property PortName : String Read FPortName Write SetPortName;
           Property Blocking : Boolean Read FBlocking Write SetBlocking;
           Property BlockingTimeout : Integer Read FBlockTime Write FBlockTime;
           Property OnConnect : TNotifyEvent Read FOnConnect Write FOnConnect;
           Property OnDisconnect : TNotifyEvent Read FOnDisconnect Write FOnDisconnect;
           Property OnRead : TNotifyReadEvent Read FOnRead Write FOnRead;
           Property OnWrite : TNotifyEvent Read FOnWrite Write FOnWrite;
     End;

     TClientSockServ = class(TClientsock)
     private
       FMustClose : Boolean;
       FServerSock : TCustomServerSock;
     protected
        Property OnConnect : TNotifyEvent;
        Property OnDisconnect : TNotifyEvent;
        Property OnRead : TNotifyReadEvent;
        Property OnWrite : TNotifyEvent;
     public
       property MustClose : Boolean read FMustClose;
       property ServerSock : TCustomServerSock read FServerSock;
     end;

     TMaxClientEvent   = procedure (Sender : TObject; Client : TClientSockServ) of object;
     TCustomServerSock = Class(TSock)
        Private
           FPortName   : String;       // Port Name Or Well-Known Number
           FSocket     : TSocket;      // Socket Handle
           FActive     : Boolean;
           FOnMaxClient: TMaxClientEvent;
           FMaxClient  : Integer;
           FBlocking   : Boolean;
           FClientList : TThreadList;
           Procedure SetPortName(Value : String);
           Procedure SetListen(Value : Boolean);
           Procedure SetBlocking(Value : Boolean);
           Function Close : Boolean;
           Procedure MyDisconnect(Sender : TObject);
        Protected
           Procedure OnSockMessage(Var Message : TMessage);override;
           Procedure Loaded; Override;
           Procedure DoAccept(Sender : TObject; NewSock : TClientSockServ);virtual;abstract;

           Procedure OnSockConnect(Sender : TObject);virtual;
           Procedure OnSockDisconnect(Sender : TObject);virtual;
           Procedure OnSockRead(Sender : TObject; Count : Integer);virtual;
           Procedure OnSockWrite(Sender : TObject);virtual;
           property Blocking : Boolean read FBlocking write SetBlocking;
        Public
           Constructor Create(AOwner : TComponent);override;
           Destructor Destroy; Override;
           function GetClientCount : Integer;
           Function PortLookup(Value : String) : U_Short;
        protected
           Property MaxClient : Integer read FMaxClient write FMaxClient;
           Property Active : Boolean read FActive write SetListen;
           Property PortName : String Read FPortName Write SetPortName;
           Property OnMaxClient : TMaxClientEvent read FOnMaxClient write FOnMaxClient;
     End;
     TBlockingServerSock = class(TCustomServerSock)
     private
           FOnAutoAccept : TNotifyAutoEvent;
     protected
           Procedure DoAccept(Sender : TObject; NewSock : TClientSockServ);override;
     public
           Constructor Create(AOwner : TComponent);override;
     published
           Property OnAutoAccept : TNotifyAutoEvent Read FOnAutoAccept Write FOnAutoAccept;
           Property Active;
           Property PortName;
           Property MaxClient;
           Property OnMaxClient;
     End;
     TNonBlockingServerSock = Class(TCustomServerSock)
     Private
           FOnConnect : TNotifyEvent;
           FOnDisconnect : TNotifyEvent;
           FOnWrite : TNotifyEvent;
           FOnRead : TNotifyReadEvent;
     Protected
           Procedure OnSockConnect(Sender : TObject);override;
           Procedure OnSockDisconnect(Sender : TObject);override;
           Procedure OnSockRead(Sender : TObject; Count : Integer);override;
           Procedure OnSockWrite(Sender : TObject);override;
     public
           Constructor Create(AOwner : TComponent);override;
     published
           Property Active;
           Property PortName;
           Property MaxClient;
           Property OnMaxClient;
           Property OnConnect : TNotifyEvent read FOnConnect write FOnConnect;
           Property OnDisconnect : TNotifyEvent read FOnDisconnect write FOnDisconnect;
           Property OnWrite : TNotifyEvent read FOnWrite write FOnWrite;
           Property OnRead : TNotifyReadEvent read FOnRead write FOnRead;
     End;

// Global IP Caching Mechanism.  Uses A String List That Stores The 32-Bit IP
// Address Of It's Associated Hostname In The Object Property Of The List.  You
// Should Never Have To Manipulate This Object Directly, But It Is Made Public
// For The Purpose Of Calling The Clear Method To Empty It.
Var IPCache : TStringList;

function LocalHostName : String; // Returns the name of this computer.
function IpAddress : string; // Returns the ipaddress of this computer.
Function WSDescription : String; // Returns A Description Of The WinSock Driver
Function WSSystemStatus : String; // Returns System Status From The WinSock Driver
Function ErrToStr(Value : Integer) : String; // Converts A WinSock Error To Text
Function Base64Encode(Value : String) : String; // Converts Passed Value To MIME Base64
Function Base64Decode(Value : String) : String; // Converts Passed Value From MIME Base64
Function URLEncode(Value : String) : String; // Converts String To A URLEncoded String
Function URLDecode(Value : String) : String; // Converts String From A URLEncoded String

Procedure Register;
Implementation


Const Base64Table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
      ValidURLChars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';

// TSockThread
Type TSockThread = Class(TThread)
        Private
           ParentSock : TCustomServerSock;
           ClientSock : TClientSockServ;
           AcSck : THandle;
           FAddr : TSockAddrIn;
        Public
           Procedure Execute; Override;
           Procedure RunThread(ParentSock : TCustomServerSock;Handle : THandle; Addr : TSockAddrIn);
     End;

Var
  WSAData : TWSAData;
  HostName : string;


Procedure TSockThread.Execute;
Begin
  ClientSock:= TClientSockserv.CreateWithSocket(ParentSock, AcSck);
  ClientSock.FServerSock:=ParentSock;
  ClientSock.FMustClose:=false;
  ClientSock.FPortName := ParentSock.fPortName;
  ClientSock.FRecvAddrIn:= FAddr;
  ParentSock.FClientList.Add(self);
  try
    ParentSock.DoAccept(ParentSock, ClientSock);
  except
  end;
  ParentSock.FClientList.Remove(self);
  ClientSock.Free;
End;


Procedure TSockThread.RunThread(ParentSock : TCustomServerSock;Handle : THandle; Addr : TSockAddrIn);
Begin
   FAddr:=Addr;
   Self.ParentSock := ParentSock;
   Self.AcSck := Handle;
//   Self.FHostName := Host;
   FreeOnTerminate := True;
   Resume;
End;
// TSock
procedure TSock.ProcessMessages;
var
  Msg : TMsg;
begin
  while PeekMessage(Msg, Handle, 0, 0, PM_REMOVE) do begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;
end;
Constructor TSock.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  hwnd:=forms.AllocateHWnd(OnSockMessage);
end;

Destructor TSock.Destroy;
begin
  DeallocateHWnd(hwnd);
  inherited Destroy;
end;

// TClientSock

Procedure TClientSock.SetHostName(Value : String);
Begin
  if Connected then
    raise ESockException.Create('Can''t change hostname while active');
  FHostName := Value;
End;

Procedure TClientSock.Abort;
var
  B : Cardinal;
begin
  b:=1;
  SetSockOpt(FSocket, SOL_SOCKET, SO_DONTLINGER, @b, sizeof(B)); // set don't linger to true 
  Close;
end;

Procedure TClientSock.SetPortName(Value : String);
Begin
  if Connected then
    raise ESockException.Create('Can''t change portname while active');
   FPortName := Value;
End;

Function TClientSock.GetText : String;
Begin
  Result := Receive;
End;

Procedure TClientSock.SetText(Value : String);
Begin
  Send(Value);
End;


Procedure TClientSock.SetBlocking(Value : Boolean);
Var Il : U_Long;
    Ev : U_Long;
Begin
   If (Not (csDesigning In ComponentState)) And (csReading In ComponentState) Then
      Begin
         // If We Haven't Fully Loaded Yet, Just Set The Value And Exit
         FBlocking := Value;
         Exit;
      End;
   If FSocket = INVALID_SOCKET Then
      FBlocking := Value
   Else
      Begin
         Ev := 0;
         FBlocking := Value;
         If FBlocking  Then
            Begin
               Il := 0;
               // Turn Off Async Checking And Set Blocking On
               WinSock.WSAAsyncSelect(FSocket, Handle, WM_SOCK, Ev);
               WinSock.IOCtlSocket(FSocket, FIONBIO, Il);
            End
         Else
            Begin
              Ev := FD_READ Or FD_CLOSE Or FD_CONNECT Or FD_WRITE Or FD_READ;
              WinSock.WSAAsyncSelect(FSocket, Handle, WM_SOCK, Ev);
            End;
      End;
  ProcessMessages;
End;

Procedure TClientSock.SetConnected(Value : Boolean);
Begin
   If Value Then
      Open
   Else
      Close;
End;

Function TClientSock.GetConnected : Boolean;
Begin
   If FSocket = INVALID_SOCKET Then
      FConnected := False;
   Result := FConnected;
End;

Function TClientSock.GetRemoteHost : String;
Begin
   Result := INet_NToA(FRecvAddrIn.SIn_Addr);
End;

Constructor TClientSock.Create(AOwner : TComponent);
Begin
   Inherited Create(AOwner);
   FHostName := '';
   FPortName := '';
   FSocket := INVALID_SOCKET;
   FInBuffer := '';
   FOutBuffer := '';
   FBlocking := False;
   FConnected := False;
End;

// This Constructor Assumes NewSocket Is A Valid Socket Handle

Constructor TClientSock.CreateWithSocket(AOwner : TComponent; NewSocket : TSocket);
Begin
   Create(nil);
   FSocket := NewSocket;
   FConnected := True;
   WinSock.WSAAsyncSelect(FSocket, Handle, WM_SOCK, FD_READ Or FD_CLOSE Or FD_CONNECT Or FD_WRITE);
End;

Destructor TClientSock.Destroy;
Begin
   If FConnected Then Close;
   Inherited Destroy;
End;

Procedure TClientSock.Loaded;
Begin
   If Not (csDesigning In ComponentState) Then
      Begin
         SetBlocking(FBlocking);
      End;
End;


Procedure TClientSock.OnSockMessage(Var Message : TMessage);
Var Event : Word;
    Error : Word;
    Res   : Integer;
Begin
   // Message Handling For Non-Blocking Sockets
   Event := WinSock.WSAGetSelectEvent(Message.LParam);
   Error := WinSock.WSAGetSelectError(Message.LParam);
   If Error > WSABASEERR Then
      exit
   Else
      // Messages Mean Different Things Depending On Whether You're Listening Or Not
      Case Event Of
         FD_CONNECT : Begin
                         FConnected := True;
                         If Assigned(FOnConnect) Then
                            FOnConnect(Self);
                      End;
         FD_CLOSE   : Begin
                         If Assigned(FOnDisconnect) Then
                            FOnDisconnect(Self);
                         Close;
                      End;
         FD_READ    : Begin
                        Res := WinSock.Recv(FSocket, FCharBuf, SizeOf(FCharBuf), 0);
                        If Res > 0 Then
                          FInBuffer := FInBuffer + Copy(FCharBuf, 1, Res);
                        If Assigned(FOnRead) Then FOnRead(Self, Length(FInBuffer));
                      End;
         FD_WRITE   : Begin
                         If FOutBuffer <> '' Then
                            Send('');
                         If Assigned(FOnWrite) Then
                            FOnWrite(Self);
                      End;
   else
     Message.Result := DefWindowProc(Handle, Message.Msg, message.Wparam, Message.Lparam);
   end;
End;

Function TClientSock.Open : Boolean;
Var Res : Integer;
Begin
   If FSocket = INVALID_SOCKET Then
      Begin
         FSocket := WinSock.Socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
         FSockAddrIn.SIn_Family := AF_INET;
         FSockAddrIn.SIn_Port := PortLookup(FPortName);
         FSockAddrIn.SIn_Addr := HostLookup(FHostName);
         SetBlocking(FBlocking);
         Res := WinSock.Connect(FSocket, FSockAddrIn, SizeOf(TSockAddrIn));
         If FBlocking Then Begin
           If Res = 0 Then Begin
             FConnected := True;
             If Assigned(FOnConnect) Then FOnConnect(Self);
           End Else Begin
             Close;
           end;
         end;
       end;
    Result := FConnected;
  ProcessMessages;
End;

Function TClientSock.Close : Boolean;
Begin
   Result := (WinSock.CloseSocket(FSocket) = 0);
   FSocket := INVALID_SOCKET;
   FConnected := False;
//  ProcessMessages;
End;

Function TClientSock.Send(Value : String) : Boolean;
Var Remain : Integer;
Begin
  Result := True;
  If FSocket = INVALID_SOCKET Then
     Raise ESockException.Create('Send - Socket Not Connected');
  FOutBuffer := FOutBuffer + Value;
  If FOutBuffer = '' Then Exit;
  If FBlocking Then Begin
    Remain := Length(FOutBuffer);
    // While Any Content Remains Or No Errors Have Happened, Then Loop
    While Remain > 0 Do Begin
      Remain := WinSock.Send(FSocket, FOutBuffer[1], Length(FOutBuffer), 0);
      If (Remain = SOCKET_ERROR) And (WinSock.WSAGetLastError <> WSAEINPROGRESS) Then Begin
        Raise ESockException.Create('Send - Socket Error');
      End Else Begin
        If Remain > 0 Then
          Delete(FOutBuffer, 1, Remain);
        Remain := Length(FOutBuffer);
      End;
    End;
    FOutBuffer := '';
  End Else Begin
     // Do Not Loop For A Non-Blocking Socket
     Remain := WinSock.Send(FSocket, FOutBuffer[1], Length(FOutBuffer), 0);
     If Remain > 0 Then Delete(FOutBuffer, 1, Remain);
  End;
  ProcessMessages;
End;

Function TClientSock.Receive : String;
Begin
   Result := ReceiveCount(-1);
End;

Function TClientSock.ReceiveCount(Count : Integer) : String;
Var Res   : Integer;
    FDSet : PFDSet;
    TV    : PTimeVal;
    Err   : Integer;
    Cnt   : Integer;
Begin
  ProcessMessages;
   If (FSocket = INVALID_SOCKET) And (FInBuffer = '') Then
      Raise ESockException.Create('Receive - Socket Not Connected');
   Cnt := Count;
   If (Cnt = -1) Or (Cnt > SizeOf(FCharBuf)) Then
      Cnt := SizeOf(FCharBuf);
   If FBlocking Then Begin
     FDSet := New(PFDSet);
     FDSet^.FD_Count := 1;
     FDSet^.FD_Array[0] := FSocket;
     If FBlockTime >= 0 Then Begin
       TV := New(PTimeVal);
       TV^.tv_sec := FBlockTime;
     End Else
       TV := Nil;
     // Used To Loop While We're Connected And Anything Is In The Input Queue
     If FConnected And (WinSock.Select(FSocket, FDSet, Nil, Nil, TV) > 0) Then Begin
       Res := WinSock.Recv(FSocket, FCharBuf, Cnt, 0);
       If (Res = SOCKET_ERROR) Then Begin
         Err := WSAGetLastError;
         Result := '';
         FInBuffer := '';
         Dispose(FDSet);
         Dispose(TV);
         Raise ESockException.Create('Receive - Socket Error '+ErrToStr(Err));
       End Else Begin
         If Res > 0 Then
           FInBuffer := FInBuffer + Copy(FCharBuf, 1, Res)
         Else
           If Res = 0 Then Begin
             Close;
             If Assigned(FOnDisconnect) Then
               FOnDisconnect(Self);
           End;
        End;
       End;
       Result := FInBuffer;
       FInBuffer := '';
       Dispose(FDSet);
       Dispose(TV);
     End Else Begin
       // non-blocking
       If((Count<>-1) And (Length(FInBuffer)>Count)) Then Begin
         Result := Copy(FInBuffer, 1, Count);
         Delete(FInBuffer, 1, Count);
       End Else Begin
         Result := FInBuffer;
         FInBuffer := '';
       End;
     End;
  ProcessMessages;
End;



Function TClientSock.HostLookup(Value : String) : TInAddr;
Type PLongInt = ^LongInt;
Var PHost : PHostEnt;
    Res   : Integer;
Begin
   If Value = '' Then
      Exit;
   FillChar(Result, SizeOf(TInAddr), #0);
   If Pos(Value[1],'0123456789') > 0 Then
      // If It's Dot-Notation, Just Convert It From An IP Address
      Result := TInAddr(WinSock.Inet_Addr(PChar(Value)))
   Else
      Begin
         Res := IPCache.IndexOf(Value);
         If Res >= 0 Then
            // It's Cached... Don't Bother Doing A Lookup
            Result.S_Addr := U_Long(IPCache.Objects[Res])
         Else
            Begin
               // Isn't Cached, Have To Do A GetHostByName
               If Value <> '' Then
                  Begin
                     PHost := WinSock.GetHostByName(PChar(Value));
                     If PHost <> Nil Then
                        Begin
                           Result.S_Addr := LongInt(PLongInt(PHost^.H_Addr_List^)^);
                           IPCache.AddObject(Value, Pointer(Result.S_Addr));
                        End
                     Else
                        Raise ESockException.Create('Host Lookup - Could Not Find Host Entry');
                  End
               Else
                  Result.S_Addr := HToNL(INADDR_ANY);
            End;
      End;
End;

Function TClientSock.PortLookup(Value : String) : U_Short;
Var PEnt : PServEnt;
Begin
   If Pos(Value[1],'0123456789') > 0 Then
      // It's Numeric, Just Convert It To A Network Byte Order Integer
      Result := HToNS(StrToInt(Value))
   Else
      Begin
         // Otherwise, Perform A GetServByName Based On The Protocol
         PEnt := WinSock.GetServByName(PChar(Value), 'tpc');
         If PEnt <> Nil Then
            Result := PEnt^.S_Port
         Else
            Raise ESockException.Create('Port Lookup - Could Not Find Service Entry');
      End;
End;

// TCustomServerSock

Procedure TCustomServerSock.SetPortName(Value : String);
Begin
   if (csLoading in Componentstate) then begin
     fportname:= value;
     exit;
   end;
   If factive Then
     raise ESockException.Create('Can''t change portname while active');
   FPortName := Value;
End;

function TCustomServerSock.GetClientCount : Integer;
begin
  with FClientlist.locklist do result:=count;
  FClientlist.Unlocklist;
end;

Function TCustomServerSock.Close : Boolean;
Begin
   Result := (WinSock.CloseSocket(FSocket) = 0);
   FSocket := INVALID_SOCKET;
   FActive := False;
  ProcessMessages;
End;

Procedure TCustomServerSock.SetBlocking(Value : Boolean);
Begin
  If (csReading In ComponentState) Then Begin
    // If We Haven't Loaded Yet, Just Set The Value And Exit
    FBlocking := Value;
    Exit;
  End;
  if fActive then Raise ESockException.Create('Blocking - Cannot change while active!');
  If (csDesigning In ComponentState) Then Begin
    Factive := Value;
    exit;
  End;
  if Value=FBlocking then exit;
  FBlocking:=Value;
End;

Procedure TCustomServerSock.SetListen(Value : Boolean);
Var WasListen : Boolean;
    Addr      : TSockAddr;
    Res       : Integer;
    I         : Integer;
    LockList  : TList;

Begin
   If (csDesigning In ComponentState) Then
      Begin
         Factive := Value;
         exit;
      End
   Else
      If (csReading In ComponentState) Then
         Begin
            // If We Haven't Loaded Yet, Just Set The Value And Exit
            FActive := Value;
            Exit;
         End;
   WasListen := FActive;
   If (FSocket <> INVALID_SOCKET) And (Not WasListen) Then
      Begin
         FActive := False;
         Raise ESockException.Create('Listen - Socket Already In Use');
      End;
   Factive := Value;
   If Factive Then
      Begin
         If Not WasListen Then
            Begin
               // Have To Create A Socket Start Asynchronous Listening
               Factive := False;
               FSocket := WinSock.Socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
               FillChar(Addr, SizeOf(Addr), #0);
               Addr.SIn_Family := AF_INET;
               Addr.SIn_Port := PortLookup(FPortName);
               Addr.SIn_Addr.S_Addr := HToNL(INADDR_ANY);
               WinSock.WSAAsyncSelect(FSocket, Handle, WM_SOCK, FD_ACCEPT);
               Res := WinSock.Bind(FSocket, Addr, SizeOf(Addr));
               If Res <> 0 Then
                  Raise ESockException.Create('Listen - Error Binding Socket');
               Res := WinSock.Listen(FSocket, 5);
               If Res <> 0 Then
                  Raise ESockException.Create('Listen - Error Starting Listen');
               Factive := True;
            End;
      End
   Else
      Begin
        locklist:=FClientList.LockList;
        for i:=0 to locklist.count-1 do begin
          TSockThread(Locklist[i]).ClientSock.Close;
        end;
        FClientlist.UnlockList;
        While GetClientCount>0 do begin
          sleep(10);
        end;
        Close;
      End;
  ProcessMessages;
End;

Constructor TCustomServerSock.Create(AOwner : TComponent);
Begin
   Inherited Create(AOwner);
   fBlocking := Blocking;
   FPortName := '';
   FSocket := INVALID_SOCKET;
   FActive := False;
   FClientList := TThreadList.Create;
End;

Destructor TCustomServerSock.Destroy;
Begin
   If FActive then
      Close;
   FClientlist.Free;
   Inherited Destroy;
End;

Procedure TCustomServerSock.Loaded;
Begin
   If Not (csDesigning In ComponentState) Then Begin
     If FActive Then Begin
       fActive:= False;
       SetListen(True);
       if FBlocking then begin
         FBlocking:=False;
         SetBlocking(True);
       end;
     End;
   End;
End;


Procedure TCustomServerSock.OnSockMessage(Var Message : TMessage);
Var Event : Word;
    Error : Word;
    AcSck : TSocket;
    Addr  : TSockAddrIn;
    AddrL : Integer;
    CSock : TClientSockserv;
    Spawn : TSockThread;
Begin
   Event := WinSock.WSAGetSelectEvent(Message.LParam);
   Error := WinSock.WSAGetSelectError(Message.LParam);
   If Error > WSABASEERR Then
     exit
   Else
      // Messages Mean Different Things Depending On Whether You're Listening Or Not
      Case Event Of
         FD_ACCEPT  : Begin
                        AddrL := SizeOf(Addr);
                        FillChar(Addr, SizeOf(Addr), #0);
   {$IFDEF VER3ORABOVE}
                        AcSck := WinSock.Accept(FSocket, @Addr, @AddrL);
   {$ELSE}
                        AcSck := WinSock.Accept(FSocket, Addr, AddrL);
   {$ENDIF}
                        if AcSck=0 then begin
                          Exit;
                        end;
//
                        if MaxClient<> 0 then begin
                          if getclientcount+1>maxclient then begin
                            if assigned(FOnMaxClient) then begin
                              CSock := TClientSockserv.CreateWithSocket(self, AcSck);
                              CSock.FRecvAddrIn := Addr;
                              fonMaxClient(self, CSock);
                              CSock.Free;
                            end else begin
                              CSock := TClientSockserv.CreateWithSocket(self, AcSck);
                              CSock.Abort;
                              CSock.Free;
                            end;
                            exit;
                          end;
                        end;
                        if FBlocking then begin
                          Spawn := TSockThread.Create(True);
                          Spawn.RunThread(Self, AcSck, Addr);
                        end else begin
                          CSock:= TClientSockserv.CreateWithSocket(Self, AcSck);
                          CSock.FServerSock:=Self;
                          CSock.FMustClose:=false;
                          CSock.FPortName := self.fPortName;
                          CSock.FRecvAddrIn:= Addr;
                          CSock.OnDisconnect:=MyDisconnect;
                          CSock.OnRead:=OnSockRead;
                          CSock.OnWrite:=OnSockWrite;
                          FClientList.Add(self);
                          OnSockConnect(CSock);
                        end;
      End;
   else
     Message.Result := DefWindowProc(Handle, Message.Msg, message.Wparam, Message.Lparam);
   end;
End;


Function TCustomServerSock.PortLookup(Value : String) : U_Short;
Var PEnt : PServEnt;
Begin
   If Pos(Value[1],'0123456789') > 0 Then
      // It's Numeric, Just Convert It To A Network Byte Order Integer
      Result := HToNS(StrToInt(Value))
   Else
      Begin
         // Otherwise, Perform A GetServByName Based On The Protocol
         PEnt := WinSock.GetServByName(PChar(Value), 'tcp');
         If PEnt <> Nil Then
            Result := PEnt^.S_Port
         Else
            Raise ESockException.Create('Port Lookup - Could Not Find Service Entry');
      End;
End;

Procedure TCustomServerSock.MyDisconnect(Sender : TObject);
begin
  FClientList.Remove(Sender);
  OnSockDisconnect(Sender);
  TSock(Sender).Free;
end;

Procedure TCustomServerSock.OnSockConnect(Sender : TObject);
begin
 // do nothing
end;

Procedure TCustomServerSock.OnSockDisconnect(Sender : TObject);
begin
 // do nothing
end;

Procedure TCustomServerSock.OnSockRead(Sender : TObject; Count : Integer);
begin
 // do nothing
end;

Procedure TCustomServerSock.OnSockWrite(Sender : TObject);
begin
 // do nothing
end;


// TBlockingServerSock
Constructor TBlockingServerSock.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  Blocking:=True;
end;

Procedure TBlockingServerSock.DoAccept(Sender : TObject; NewSock : TClientSockserv);
begin
  if assigned(fonAutoAccept) then
    fonAutoAccept(sender, newsock);
end;

// TNonBlockingServerSock
Procedure TNonBlockingServerSock.OnSockConnect(Sender : TObject);
begin
  if Assigned(FOnConnect) then
  FOnConnect(Sender);
end;

Procedure TNonBlockingServerSock.OnSockDisconnect(Sender : TObject);
begin
  if assigned(FOnDisconnect) then
  FOnDisconnect(Sender);
end;

Procedure TNonBlockingServerSock.OnSockRead(Sender : TObject; Count : Integer);
begin
  if Assigned(FOnRead) then
  FOnRead(Sender, Count);
end;
Procedure TNonBlockingServerSock.OnSockWrite(Sender : TObject);
begin
  if Assigned(Fonwrite) then
  FOnWrite(Sender);
end;

Constructor TNonBlockingServerSock.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  Blocking:=False;
end;

// Misc
function LocalHostName : String;
begin
  result:=HostName;
end;

Function WSDescription : String;
Begin
   Result := StrPas(WSAData.szDescription);
End;

Function WSSystemStatus : String;
Begin
   Result := StrPas(WSAData.szSystemStatus);
End;


Function ErrToStr(Value : Integer) : String;
Begin
   Result := 'UNKNOWN ERROR';
   Case Value Of
      WSABASEERR+4 : Result := 'WSAEINTR';
      WSABASEERR+9 : Result := 'WSAEBADF';
      WSABASEERR+13 : Result := 'WSAEACCES';
      WSABASEERR+14 : Result := 'WSAEFAULT';
      WSABASEERR+22 : Result := 'WSAEINVAL';
      WSABASEERR+24 : Result := 'WSAEMFILE';
      WSABASEERR+35 : Result := 'WSAEWOULDBLOCK';
      WSABASEERR+36 : Result := 'WSAEINPROGRESS';
      WSABASEERR+37 : Result := 'WSAEALREADY';
      WSABASEERR+38 : Result := 'WSAENOTSOCK';
      WSABASEERR+39 : Result := 'WSAEDESTADDRREQ';
      WSABASEERR+40 : Result := 'WSAEMSGSIZE';
      WSABASEERR+41 : Result := 'WSAEPROTOTYPE';
      WSABASEERR+42 : Result := 'WSAENOPROTOOPT';
      WSABASEERR+43 : Result := 'WSAEPROTONOSUPPORT';
      WSABASEERR+44 : Result := 'WSAESOCKTNOSUPPORT';
      WSABASEERR+45 : Result := 'WSAEOPNOTSUPP';
      WSABASEERR+46 : Result := 'WSAEPFNOSUPPORT';
      WSABASEERR+47 : Result := 'WSAEAFNOSUPPORT';
      WSABASEERR+48 : Result := 'WSAEADDRINUSE';
      WSABASEERR+49 : Result := 'WSAEADDRNOTAVAIL';
      WSABASEERR+50 : Result := 'WSAENETDOWN';
      WSABASEERR+51 : Result := 'WSAENETUNREACH';
      WSABASEERR+52 : Result := 'WSAENETRESET';
      WSABASEERR+53 : Result := 'WSAECONNABORTED';
      WSABASEERR+54 : Result := 'WSAECONNRESET';
      WSABASEERR+55 : Result := 'WSAENOBUFS';
      WSABASEERR+56 : Result := 'WSAEISCONN';
      WSABASEERR+57 : Result := 'WSAENOTCONN';
      WSABASEERR+58 : Result := 'WSAESHUTDOWN';
      WSABASEERR+59 : Result := 'WSAETOOMANYREFS';
      WSABASEERR+60 : Result := 'WSAETIMEDOUT';
      WSABASEERR+61 : Result := 'WSAECONNREFUSED';
      WSABASEERR+62 : Result := 'WSAELOOP';
      WSABASEERR+63 : Result := 'WSAENAMETOOLONG';
      WSABASEERR+64 : Result := 'WSAEHOSTDOWN';
      WSABASEERR+65 : Result := 'WSAEHOSTUNREACH';
      WSABASEERR+66 : Result := 'WSAENOTEMPTY';
      WSABASEERR+67 : Result := 'WSAEPROCLIM';
      WSABASEERR+68 : Result := 'WSAEUSERS';
      WSABASEERR+69 : Result := 'WSAEDQUOT';
      WSABASEERR+70 : Result := 'WSAESTALE';
      WSABASEERR+71 : Result := 'WSAEREMOTE';
      WSABASEERR+91 : Result := 'WSASYSNOTREADY';
      WSABASEERR+92 : Result := 'WSAVERNOTSUPPORTED';
      WSABASEERR+93 : Result := 'WSANOTINITIALISED';
      WSABASEERR+101 : Result := 'WSAEDISCON';
      WSABASEERR+1001 : Result := 'WSAHOST_NOT_FOUND';
      WSABASEERR+1002 : Result := 'WSATRY_AGAIN';
      WSABASEERR+1003 : Result := 'WSANO_RECOVERY';
      WSABASEERR+1004 : Result := 'WSANO_DATA';
   End;
End;

Function Base64Encode(Value : String) : String;
Var AIn   : Array[1..3] Of Byte;
    AOut  : Array[1..4] Of Byte;
    AWork : Array[1..3] Of Byte;
    I : Integer;
    O : LongInt;
Begin
   Result := '';
   I := 1;
   O := Length(Value);
   Case Length(Value) Mod 3 Of
      1 : Value := Value + #0 + #0;
      2 : Value := Value + #0;
   End;
   While I < Length(Value) Do
      Begin
         AIn[1] := Byte(Value[I]);
         AIn[2] := Byte(Value[I+1]);
         AIn[3] := Byte(Value[I+2]);

         AOut[1] := Byte(AIn[1] Shr 2);
         AWork[1] := Byte(AIn[1] Shl 4);
         AWork[2] := Byte(AWork[1] And $30);
         AWork[3] := Byte(AIn[2] Shr 4);
         AOut[2] := Byte(AWork[2] Or AWork[3]);
         AWork[1] := Byte(AIn[2] Shl 2);
         AWork[2] := Byte(AWork[1] And $3C);
         AWork[3] := Byte(AIn[3] Shr 6);
         AOut[3] := Byte(AWork[2] Or AWork[3]);
         AOut[4] := Byte(AIn[3] And $3F);

         Inc(I, 3);
         Result := Result + Base64Table[AOut[1]+1] + Base64Table[AOut[2]+1] + Base64Table[AOut[3]+1] + Base64Table[AOut[4]+1];
      End;
   If O Mod 3 > 0 Then
      Result[Length(Result)] := '=';
   If O Mod 3 = 1 Then
      Result[Length(Result)-1] := '=';
End;

Function Base64Decode(Value : String) : String;
Var AIn   : Array[1..4] Of Byte;
    AOut  : Array[1..3] Of Byte;
    AWork : Array[1..3] Of Byte;
    I : Integer;
    C : Integer;
Begin
   Result := '';
   I := 1;
   While I < Length(Value) Do
      Begin
         C := 3;
         FillChar(AWork, SizeOf(AWork), #0);
         FillChar(AOut, SizeOf(AWork), #0);
         AIn[1] := Byte(Pos(Value[I],Base64Table)-1);
         AIn[2] := Byte(Pos(Value[I+1],Base64Table)-1);
         AIn[3] := Byte(Pos(Value[I+2],Base64Table)-1);
         AIn[4] := Byte(Pos(Value[I+3],Base64Table)-1);
         If Value[I+3]='=' Then
            Begin
               C := 2;
               AIn[4] := 0;
               If Value[I+2]='=' Then
                  Begin
                     C := 1;
                     AIn[3] := 0;
                  End;
            End;
         AWork[2] := Byte(AIn[1] Shl 2);
         AWork[3] := Byte(AIn[2] Shr 4);
         AOut[1] := Byte(AWork[2] Or AWork[3]);
         AWork[2] := Byte(AIn[2] Shl 4);
         AWork[3] := Byte(AIn[3] Shr 2);
         AOut[2] := Byte(AWork[2] Or AWork[3]);
         AWork[2] := Byte(AIn[3] Shl 6);
         AOut[3] := Byte(AWork[2] Or AIn[4]);
         Result := Result + Char(AOut[1]);
         If C > 1 Then
            Result := Result + Char(AOut[2]);
         If C > 2 Then
            Result := Result + Char(AOut[3]);
         Inc(I, 4);
      End;
End;

Function URLEncode(Value : String) : String;
Var I : Integer;
Begin
   Result := '';
   For I := 1 To Length(Value) Do
      Begin
         If Pos(UpperCase(Value[I]), ValidURLChars) > 0 Then
            Result := Result + Value[I]
         Else
            Begin
               If Value[I] = ' ' Then
                  Result := Result + '+'
               Else
                  Begin
                     Result := Result + '%';
                     Result := Result + IntToHex(Byte(Value[I]), 2);
                  End;
            End;
      End;
End;

Function URLDecode(Value : String) : String;
Const HexChars = '0123456789ABCDEF';
Var I        : Integer;
    Ch,H1,H2 : Char;
Begin
   Result := '';
   I := 1;
   While I <= Length(Value) Do
      Begin
         Ch := Value[I];
         Case Ch Of
            '%' : Begin
                     H1 := Value[I+1];
                     H2 := Value[I+2];
                     Inc(I, 2);
                     Result := Result + Chr(((Pos(H1, HexChars) - 1) * 16) + (Pos(H2, HexChars) - 1));
                  End;
            '+' : Result := Result + ' ';
            '&' : Result := Result + #13+#10;
            Else Result := Result + Ch;
         End;
         Inc(I);
      End;
End;

Procedure Register;
Begin
  RegisterComponents('TSock', [TClientSock, TBlockingServerSock, TNonBlockingServerSock]);
End;

function IpAddress : string;
var
  p: pHostEnt;
  T : PInAddr;
begin
  p:=gethostbyname(pchar(hostname));
  if p=nil then begin
    result:='127.0.0.1';
    exit;
  end;
  t:=PInAddr(p.h_addr_list^);
  if t<>nil then begin
    result:=StrPas(inet_ntoa(t^));
  end;
end;

var
  Buffer : array [1..64] of char;

Initialization
   If WinSock.WSAStartup($0101, WSAData) <> 0 Then
      Raise ESockException.Create('WSAStartup - Could Not Initialize WinSock');
   IPCache := TStringList.Create;
  if gethostname(@Buffer, SizeOf(Buffer)) <> 0 then
    HostName:='127.0.0.1'
  else
    HostName:= StrPas(@Buffer);

Finalization
   IPCache.Free;
   WinSock.WSACleanup;
End.
